/*
 *  Copyright (c) 2000 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 */


#include "NetLogger.h"

#include <stdexcept>
#include <sys/time.h>

using std::string;

/* Constructor */
NetLogger::NetLogger()
{
   mHandle_p = NULL;
}

/* Open for output to URL 
 * see NetLoggerOpen() in C API.
 */
void
NetLogger::open (const string& progName, const string& url, 
		 int flags)
{
    mHandle_p = NetLoggerOpen( const_cast<char*>(progName.c_str()),
			       const_cast<char*>(url.c_str()),
			       flags );
    if (mHandle_p == NULL)
    {
        throw runtime_error ("Unable to open log file");
    }
}

/* Open for output to a file descriptor 
 */
void 
NetLogger::open(const std::string& progName, int fd, int flags )
{
    mHandle_p = NetLoggerOpenFd( const_cast<char*>(progName.c_str()),
			       fd,
			       flags );
    if (mHandle_p == NULL)
    {
        throw runtime_error ("Unable to open log file");
    }
}

/* Set the logging level. Level 0, the default, means "write all".
 * If !=0, write()s with a level that is greater than this level 
 * will return immediately.
 */
void 
NetLogger::setLevel(int lvl)
{
    NetLoggerSetLevel( mHandle_p, lvl );
}

/* Write a NetLogger message */   
void
NetLogger::write (const string& keyword, const string& mesg)
{
    struct timeval tv;
    gettimeofday (&tv, 0);
    if (NetLoggerGTWriteString (mHandle_p, tv.tv_sec, tv.tv_usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str())) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Write a NetLogger message with user-provided timestamp */
void
NetLogger::write (const string& keyword, const string& mesg, long sec, long usec)
{
    if (NetLoggerGTWriteString (mHandle_p, sec, usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str()) ) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Write a NetLogger message of a certain 'level' */   
void 
NetLogger::write(int lvl,
		 const std::string& keyword,
		 const std::string& mesg)
{
    struct timeval tv;
    gettimeofday (&tv, 0);
    if (NetLoggerGTWriteStringLvl (mHandle_p, lvl, tv.tv_sec, tv.tv_usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str())) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}
    
/* Write a NetLogger message of a certain 'level' 
 * with user-provided timestamp */
void 
NetLogger::write(int lvl, const std::string& keyword, 
		 const std::string& mesg, 
		 long sec, long usec)
{
    if (NetLoggerGTWriteStringLvl (mHandle_p, lvl, sec, usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str()) ) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Flush buffered messages 
 * Return: 1=something was flushed; 0=no msgs flushed
 */
int
NetLogger::flush()
{
   return NetLoggerFlush( mHandle_p );
}

/* Set a 'tag' that is appended to all msgs --
 * for all instances of NetLogger!! */
void NetLogger::setTag( const string& tag )
{
   NetLoggerSetTag( const_cast<char *>(tag.c_str()) );
}

/* Flush buffer, close socket/file/syslog */
void
NetLogger::close ()
{
    if (NetLoggerClose (mHandle_p) == -1)
	throw runtime_error ("Unable to close the log");
}
