/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley 
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 * 
 *  $Id: securetcp.cc,v 1.26 2001/10/15 21:23:46 boverhof Exp $
 */

/* ---------------------------------------------------------
 *  SecureTcp:
 *   	This class defines both the client side and the serverside
 *  	of an authenticated and authorized socket connection.  It
 *  	uses SSL to provide the secure channel and Akenti certificates
 *  	to provide the authentication and authorization.
 *      Authors: Deb Agarwal, Keith Jackson, and Marcia Perry 
 *
 * --------------------------------------------------------- */

#include "securetcp.h"
#include "openssl/rand.h"
#include <iostream>
#include <cstring>
#include <string>
#include <sys/stat.h>

using std::cerr;
using std::endl;
using std::string;


/* ---------------------------------------------------------
 *   Constructor for the secure tcp object.  The arguments are
 *   all given defaults in the header file declaration so this
 *   constructor matches any number of arguments.  The arguments are
 *      serv - 1 or 0 indicating whether this is a server (1=server)
 *      fd   - file descriptor
 *      timeout - timeout
 *      p    - port
 *   This constructor calls the constructor for the Network superclass.
 *
 * --------------------------------------------------------- */

SecureTcp::SecureTcp( string cert, string key, string ca )
    : MyDebug("SecureTcp: "), func( NULL ), 
      mCertfilename( cert ), mKeyfilename( key ), mCAfilename( ca )
{
    sock_p = NULL;
}

SecureTcp::~SecureTcp()
{
    printDebug(5, "~SecureTCP\n");

    // if server, sock_p maintained elsewhere in NetworkTCP
    if( (sock_p!=NULL) && !mIsServer ){
	delete sock_p;
	sock_p = NULL;
    }
    else
    	closeSocket();

}


/* ---------------------------------------------------------
 *  setPassword_cb: Optional. Sets a function pointer to a 
 *	function which is called when SSL requests a password. 
 *
 * --------------------------------------------------------- */

void SecureTcp::setPassword_cb( CallBackPtr funcPtr )
{
    func =  funcPtr;
}


/* ---------------------------------------------------------
 * init: 
 *
 * --------------------------------------------------------- */

bool SecureTcp::init( bool serv, const string host )
{
    int verify;
    mIsServer = serv;

    printDebug( 3, "In init and mIsServer = %d\n", mIsServer );
    if( mIsServer )
      verify = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
    else
      verify = SSL_VERIFY_PEER;

    SSL_load_error_strings();
    printDebug( 3, "Set the verify variable\n");
    SSLeay_add_ssl_algorithms();
    printDebug( 3, "Finished setting ssl algorithms\n");

    if( mIsServer )
      mMeth_p = SSLv23_server_method();
    else{
      mMeth_p = SSLv3_client_method();
    }

    mCtx_p = NULL;
    mCtx_p = SSL_CTX_new(mMeth_p);
    if (!mCtx_p) {
      ERR_print_errors_fp(stderr);
      //exit(2);
      return false;
    }

    printDebug( 3, "Certificate file name is %s\n", mCertfilename.c_str() );
    if (SSL_CTX_use_certificate_file(mCtx_p, mCertfilename.c_str(), 
				               SSL_FILETYPE_PEM) <= 0) {
      ERR_print_errors_fp(stderr);
      return false;
    }

    SSL_CTX_set_default_passwd_cb( mCtx_p, func );
    printDebug( 3, "key file name is %s\n", mKeyfilename.c_str() );
    if (SSL_CTX_use_PrivateKey_file(mCtx_p, mKeyfilename.c_str(), 
				               SSL_FILETYPE_PEM) <= 0) {
      ERR_print_errors_fp(stderr);
      return false;
    }

    if (!SSL_CTX_check_private_key(mCtx_p)) {
      fprintf(stderr,"Private key does not match the certificate public key\n");
      return false;
    }
    SSL_CTX_set_verify(mCtx_p, verify, NULL);
    SSL_CTX_load_verify_locations(mCtx_p, mCAfilename.c_str(), NULL);


    if( mIsServer )
      SSL_CTX_set_client_CA_list (mCtx_p, SSL_load_client_CA_file (mCAfilename.c_str()));
    mRemoteHost = host;
    printDebug( 3, "At the end of init remoteHost = %s\n", 
		                               mRemoteHost.c_str());
    return true;
}


/* ---------------------------------------------------------
 *  doConnect:
 *	TCPNet sock_p  - locally maintained only for connect.
 *
 * --------------------------------------------------------- */

int SecureTcp::doConnect( SecurityInfo* secInfo, const string hostname )
{
    /* This routine should only ever be called by the client 
     * side of the connection (the server side calls doAccept).
     */
    if( mIsServer ) {
	cerr << "ERROR: doConnect called by the server side of the connection\n";
	return (-1);
    }
    if( sock_p != NULL ) {
	cerr << "ERROR: doConnect socket already allocated\n";
	return (-1);
    }
	
    int err;

    sock_p = new TCPNet();
    string testhost = hostname;
    if ( sock_p == NULL || (sock_p->openSocket(false) < 0) ) {
        cerr << "sock_p.openSocket() failed\n";
        delete sock_p;
	sock_p = NULL;
        return (-1);
    }

    if( (sock_p->Connect(testhost)) != 0 ) {
        cerr << "sock_p.Connect(" << testhost << ") failed\n";
        delete sock_p;
	sock_p = NULL;
        return (-1);
    }

    setRandFile(0);
    mCon_p = NULL;

    printDebug( 3, "doConnect--calling SSL_new(mCtx_p)\n");
    mCon_p = (SSL *)SSL_new(mCtx_p);
    printDebug( 3, "doConnect--SSL_new(mCtx_p) returned mCon_p = %p\n", mCon_p);

    if (!mCon_p) {  	
	printf("doConnect: mCon_p is null\n");
	return (-1);
    }

    mCon_p->debug=1;
    printDebug( 3, "doConnect()--calling SSL_set_fd() xSock=%d\n", sock_p->getSock() );

    err = SSL_set_fd(mCon_p, sock_p->getSock() );
    if( err < 0 ) {
        perror("SSL_set_fd");
	return (-1);
    }
    printDebug( 3, "doConnect()--calling SSL_connect()\n");

    err = SSL_connect(mCon_p);
    printDebug(3, "**after connect err=%d\n", err);

    if(err<=0) {
	int x = SSL_get_error(mCon_p, err);
        if( x == SSL_ERROR_SSL )
	    printf("SSL_ERROR_SSL\n");
  	else
	    printf("not SSL_ERROR_SSL\n");

	char* errstr = ERR_error_string( ERR_get_error(), NULL );
	printf("ERROR: %s\n", errstr);
	return (-1);
    }

    if( err < 0 ){
	perror("SSL_connect");
	return (-1);
    }

    printDebug( 3, "doConnect()-- SSL_connect() returned err %d\n", err);
    printDebug( 3, "SSL connection using %s\n", SSL_get_cipher(mCon_p));

    return 0;
}


/* ---------------------------------------------------------
 *  doAccept
 * 	TCPNet *mySock_p - maintained in NetworkTCP
 * --------------------------------------------------------- */

int SecureTcp::doAccept( SecurityInfo* secInfo, TCPNet *mySock_p )
{
  /* if the caller did not supply a timeout assume that they
   * want a short default timeout value.
   */

    struct timeval timeout;
    int    timesecs = 1;

    timeout.tv_sec = timesecs;
    timeout.tv_usec = 0;
    return doAccept( secInfo, &timeout, mySock_p );
}

int SecureTcp::doBlockedAccept( SecurityInfo* secInfo, TCPNet *mySock_p )
{
  /* This routine is called when the caller does not.
   * want the routine to return until after someone has connected.
   */	

  printDebug( 3, "called blocked accept\n");
  return doAccept( secInfo, (struct timeval *)NULL, mySock_p );
  
}

int SecureTcp::doAccept(SecurityInfo* secInfo, struct timeval *t, TCPNet *mySock_p )
{
    if( !mIsServer || mySock_p==NULL )
	return -1;

    fd_set fd;
    int n, err;
    char *str;
    char *sslDN; // Distinguished name as represented by SSL

    sock_p = mySock_p;

    if( t != NULL )
        printDebug( 3, "doAccept()--starting with timeout: %d sec %d usec\n",t->tv_sec,
		 t->tv_usec);
    else
        printDebug( 3, "Called accept with infinite timeout\n");

    X509 *peerCert;
    FD_ZERO(&fd);
    FD_SET(sock_p->getSock(), &fd);
    printDebug( 3, "doAccept()--calling  select()\n");
    n = select( (sock_p->getSock())+1, &fd, NULL, NULL, t );
    printDebug( 3, "doAccept--select() returned %d\n", n);

    if( n < 0 )
    {
        printf("ERROR: SecureTcp::doAccept--select returned error\n");
	return n;
    }
    else if (n > 0)
    {
	//int len = sizeof(peer);
        printDebug(3, "doAccept--calling accept()\n");

	//mNewsock = mySock.Accept(); 
	mNewsock = sock_p->getSock(); 
	printDebug(3, "mNewsock= %d\n", mNewsock);

	if( mNewsock < 0 )
	{
	    printError("SecureTcp--sock_p->Accept( peer )"); 	
	    return (-1);
	}
	
        printDebug( 3, "doAccept--calling SSL_new(mCtx_p)\n");
        mCon_p = (SSL *)SSL_new(mCtx_p);
	printDebug( 3, "doAccept--SSL_new(mCtx_p) returned mCon_p = %p\n", mCon_p);
        if (!mCon_p)
	{  	
	    printf("doAccept: mCon_p is null\n");
	    return (-1);
  	}
	mCon_p->debug=1;

        printDebug( 3, "doAccept()-- SSL_set_fd()\n");
        SSL_set_fd(mCon_p, mNewsock);

        printDebug( 3, "doAccept()-- SSL_accept()\n");
	err = SSL_accept ( mCon_p );
	if( err<0 )
	{
    	    ERR_print_errors_fp(stderr);
	    printf("SSL_accept: error\n");
	    return (-1);
        }

	printDebug( 3, "SSL connection using %s\n", SSL_get_cipher (mCon_p));
        peerCert = SSL_get_peer_certificate(mCon_p);
        if( peerCert == NULL)
	{
	    printError("didn't get the peer certificate\n");
	    // no cert we should drop this connection
	    //close(sock);
	    return(-1);
	}
        else
	{
	    printDebug( 3, "Client certificate:\n");
	    sslDN = X509_NAME_oneline (X509_get_subject_name (peerCert), 0, 0);
	    if(sslDN==NULL){
	        printf("X509_NAME_oneline error: subject name\n");
		return (-1);
	    }
	    printDebug( 3, "\t subject: %s\n", sslDN );
	    str = X509_NAME_oneline (X509_get_issuer_name  (peerCert), 0, 0);
	    if(str==NULL){
	        printf("X509_NAME_oneline error: issuer name\n");
		return (-1);
	    }
	    printDebug( 3, "\t issuer: %s\n", str);
	    // Use the DN to check the ACL file.
	    // If the requestor isn't authorized for any tests,
	    // drop the connection.  Otherwise, fill in the local
	    // authorization cache.

	    printDebug(3, "securetcp  sslDN: %s  str: %s\n", sslDN, str);

	    try
	    {
		const string aclfile("ACLFile");
		Authorizer* auth = new ACLAuthorizer(aclfile);

	  	printDebug(3, "Authorizer returned\n");
		string execute("execute");
	  	bool ret = auth->authorize(sslDN, execute.c_str());
		delete auth;

	  	printDebug(3, "Return is: %d\n", ret);
		if(ret!=1)
		{
		    printf("ERROR: SecureTcp::doAccept--ACL authorization failed\n");
		    return -1;
		}
	    }
	    catch(string& ex)
	    {
		cerr << "ERROR: SecureTcp::doAccept--ACL authorization failed\n";
		cerr << "Exception: " << ex << endl;
		return -1;
	    }

	    free (str);
	    X509_free (peerCert);
	} 
	// allocate a string to hold the version of the DN used to
	// hash in security info.
        char* nDN = new char[strlen(sslDN)];
        parseDN(nDN, sslDN);
        SSL_SESSION* session = SSL_get_session(mCon_p);
        unsigned char* ms = session->master_key;
        printDebug( 3, "the length of the ms is: %d\n", (int)(strlen((char*)ms)));
        int l = strlen((char*)ms);
        char* buf = new char[l];
        string stMS;
        for (int i=0; i < l; i++)
        {
	    sprintf(buf, "%02X", ms[i]);
            stMS += buf;
        }
        delete[] buf;
	free( sslDN );
	printDebug( 3, "about to call setData, MS is: %s\n", stMS.c_str() );
	secInfo->setData(nDN, stMS.c_str()); 
	delete[] nDN;
    }
    printDebug( 3, "doAccept()--returning %d\n", n );
    return n;
}


/* ---------------------------------------------------------
 *  closeClientSock()
 *
 * --------------------------------------------------------- */

void SecureTcp::closeClientSock()
{
  if( mNewsock > 0 )
     close( mNewsock );
  mNewsock = 0;
}


/* ---------------------------------------------------------
 * A method to extract the C, O, OU, and CN from the DN.  It takes a / seperated
 * DN, and returns the concatanation of the C, O, OU, and CN.
 * @param newDN a char* to place the truncated DN into.
 * @param oldDN a char* containing the original DN.
 *
 * It assumes that the DN is always passed in in the order of C, O, OU, CN.
 *  This routine parses the distinguished name and gets rid of all
 *  the fields except the ones it is looking for.  This new DN will be
 *  used to hash on in SecurityInfo structure . . .
 *
 * --------------------------------------------------------- */

int SecureTcp::parseDN(char* newDN, const char* oldDN) 
{
    newDN[0] = '\0';
    char* tmpDN = strdup(oldDN); 
    char* tokenPtr = strtok(tmpDN, "/");
    if (tokenPtr == NULL) {
    	cerr << "unable to parse the DN" << endl;
    	free(tmpDN);
    	return (-1);
    }

    while (tokenPtr != NULL) {
      	printDebug( 3, "tokenPtr is: %s\n", tokenPtr );
      	printDebug( 3, "tokenPtr[0] is: %c\n", tokenPtr[0] );
      	if ((tokenPtr[0] == 'C') || (tokenPtr[0] == 'O')) {
	    if (tokenPtr[1] == '=') {
	    	strcat(newDN, tokenPtr);
	      	printDebug( 3, "newDN is now: %s\n", newDN);
	    }
	    else {
	      	if ((tokenPtr[1] == 'U') || (tokenPtr[1] == 'N')) {
		    if (tokenPtr[2] == '=') {
		    	strcat(newDN, tokenPtr);
		      	printDebug( 3, "newDN is now: %s\n", newDN);
		    }
	        }
	    } 
        }
        tokenPtr = strtok( NULL, "/");
    }
    free(tmpDN);
    printDebug( 3, "about to leave parseDN, and newDN is: %s\n", newDN);
    return 1;
}


/* --------------------------------------------------------- 
 *  writen:
 *  	A method to write to the secure socket.  It accepts a
 *  	character buffer and sends it.  
 *
 * --------------------------------------------------------- */

void SecureTcp::writen( char *msg, int len ) 
{
    printDebug( 3, "writen\n");
    SSL_write( mCon_p, msg, len);
}


/* --------------------------------------------------------- 
 *  readn:
 *  	A method to read from the secure socket.  It assumes
 *  	that the message comes in as a character buffer and
 *  	fills in the passed in buffer.
 *
 * --------------------------------------------------------- */

void SecureTcp::readn( char *buf, int len ) 
{
    printDebug( 3, "readn\n");
    SSL_read( mCon_p, buf, len);
}


/* --------------------------------------------------------- 
 *  closeSocket:
 *
 * --------------------------------------------------------- */

void SecureTcp::closeSocket()
{
    if( sock_p != NULL )
    	sock_p->CloseClient(); 
    if( !mCon_p )
      	SSL_free( mCon_p );
    if( !mCtx_p )
      	SSL_CTX_free( mCtx_p);
}


/* ---------------------------------------------------------------------
 *  setRandFile:
 *
 *	Sets up the error tables used by SSL and adds ours
 *  	using the ERR_LIB_USER
 *  	Only the first call does anything.
 *      Will also add any builtin objects for SSLeay.
 *
 * --------------------------------------------------------------------*/

int SecureTcp::setRandFile(int i)
{
        static int init=1;

    struct stat stx;
    clock_t cputime;

#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
    const char *randfile;
#else
    char *randfile;
#endif

    char *egd_path;
    char buffer[200];
    printDebug(3,"ERR_load_prxyerr_strings, init=%d,i=%d\n",init,i);

    if (init) {
        init=0;
#ifndef RAND_DO_NOT_USE_CLOCK
                clock();
#endif
                if (i == 0) {
                    SSL_load_error_strings();
                }
                OBJ_create("1.3.6.1.4.1.3536.1.1.1.1","CLASSADD","ClassAdd");
                OBJ_create("0.9.2342.19200300.100.1.1","USERID","userId");

                //ERR_load_strings(ERR_LIB_USER+i,prxyerr_str_functs);
                //ERR_load_strings(ERR_LIB_USER+i,prxyerr_str_reasons);
                //ERR_user_lib_prxyerr_number = ERR_LIB_USER+i;
                /*
                 * We need to get a lot of randomness for good security
                 * OpenSSL will use /dev/urandom (if available),
                 * uid, time, and gid.
                 *
                 * If user has RANDFILE set, or $HOME/.rnd
                 * load it for extra random seed.
                 * This may also not be enough, so we will also add in
                 * the time it takes to run this routine, which includes
                 * reading the randfile.
                 * Later we will also add in some keys and some stats
                 * if we have them.
                 * look for RAND_add in this source file.
                 *
                 * Other methods we could use:
                 *  * Librand from  Don Mitchell and Matt Blaze
                 *  * Doing a netstat -in
                 *  * some form of pstat
                 * But /dev/random and/or egd should be enough.
                 */

                randfile=RAND_file_name(buffer,200);
                if (randfile) {
                    RAND_load_file(randfile,1024L*1024L);
                }

#if SSLEAY_VERSION_NUMBER >=  0x0090581fL
                /*
                 * Try to use the Entropy Garthering Deamon
                 * See the OpenSSL crypto/rand/rand_egd.c
                 */
                egd_path = getenv("EGD_PATH");
                if (egd_path == NULL) {
                    egd_path = "/etc/entropy";
                }
                RAND_egd(egd_path);
#endif

                /* if still not enough entropy*/
                if (RAND_status() == 0) {
                        stat("/tmp",&stx); /* get times /tmp was modified */
                        RAND_add((void*)&stx,sizeof(stx),16);
                }

#ifndef RAND_DO_NOT_USE_CLOCK
                cputime=clock();
                RAND_add((void*)&cputime, sizeof(cputime),8);
#endif

                printDebug(3,"ERR_load_prxyerr_strings RAND_status=%d,i=%d\n", RAND_status(),i);
                i++;
#ifdef USE_PKCS11
        i = ERR_load_scerr_strings(i);
#endif

    }
        return i;
}


/* --------------------------------------------------------------------- 
 *  getLocalName:
 *
 * --------------------------------------------------------------------- */

string
SecureTcp::getLocalName( void )
{
    return *(sock_p->getLocalName());
}


/* ---------------------------------------------------------------------
 *  getRemoteName:
 *
 * --------------------------------------------------------------------- */

string
SecureTcp::getRemoteName( void )
{
    return *(sock_p->getRemoteHost());
}
