/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 *  $Id: tcpnet.cc,v 1.13 2001/08/31 23:08:57 dang Exp $
 */

#include "tcpnet.h"
#include "headers.h"
#include <iostream>
#include <cstring>
#include <string>
#include <errno.h>

using std::string;
using std::cout;
using std::endl;


/* ---------------------------------------------------------
 * TCPNet:
 *	 stores server port and TCP mode
 *
 * --------------------------------------------------------- */

TCPNet::TCPNet( int fd, int timeout, short inPort): Network(fd,timeout,inPort), 
        xSock(-1), xPort(inPort), isServer(false), xRemoteHost(string(""))
{
}

TCPNet::~TCPNet()
{
    if( xSock > 0 ) 
	close( xSock );
}


/* ---------------------------------------------------------
 * openSocket:
 *
 * --------------------------------------------------------- */

int
TCPNet::openSocket( bool isServ )
{
    printDebug(4, "openSocket\n");
    int i = 1;
    isServer = isServ;

    //  fill in the sockaddr_in struct with the appropriate values
    memset (&peer, '\0', sizeof(peer));
    peer.sin_family = AF_INET;
    printDebug( 4, "setting port %d\n", (int)peer.sin_port );
    peer.sin_port = htons(port);

    if( isServer ){
	printDebug(4, "Server\n");
	peer.sin_addr.s_addr = INADDR_ANY;
    }
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0){
	printError("openSocket--socket() returned %d\n", sock );
	return (-1);
    }
    if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&i,sizeof(i)) < 0){
	perror("setsockopt()");
	close(sock);
	return (-1);
    }
    printDebug(4, "return from openSocket  sock=%d\n", sock);
    return 0;
}


/* ---------------------------------------------------------
 * Listen:
 *  	server bind and listen
 *
 * --------------------------------------------------------- */ 

int 
TCPNet::Listen( void )
{
    printDebug(4, "Listen() bind\n");
    if ( bind(sock,(struct sockaddr *)&peer,sizeof(peer)) < 0 ){
       printError( "Socket--bind()");
       close(sock);
       return (-1);
    }  
    printDebug(4, "Listen() listen\n");
    if ( listen(sock,5) < 0 ){
	printError( "Socket--listen()");
	close(sock);
	return (-1);
    }
    printDebug(4, "return from Listen  sock=%d\n", sock);
    return 0;
}


/* ---------------------------------------------------------
 * Accept:
 *
 * --------------------------------------------------------- */

int 
TCPNet::Accept( void )
{
    printDebug(4, "Accept()\n");
    int len = sizeof(peer);

    xSock   = accept(sock, (struct sockaddr *)&peer, (Socklen_t*)&len);
    if( xSock < 0 ){
	printError( "Socket--accept()" );
	close( sock );
	return -1;
    }
    printDebug(4, "return from TCPNet::Accept()\n");
    return xSock;
}


/* ---------------------------------------------------------
 * Connect:
 *
 * --------------------------------------------------------- */

int
TCPNet::Connect( string test_server )
{
    peer.sin_addr.s_addr = convertAddr( test_server.c_str() );
//    char* temp= (char*)peer.sin_addr.s_addr;
//    printDebug(4, "Client to connect to %u.%u.%u.%u\n",
//		temp[0],temp[1],temp[2],temp[3] );

    xSock = sock;
    printDebug(4, "TCPNet::Connect(%d)\n", xSock);
    if( connect( xSock, (struct sockaddr *)&peer, sizeof(peer)) < 0 ){
	printError( "Socket--connect()" );
	close( xSock );
        return -1;
    }
    return 0;
}


/* ---------------------------------------------------------
 * CloseServer:
 *
 * --------------------------------------------------------- */

void 
TCPNet::CloseServer( void )
{
    printDebug(4, "CloseServer() xSock:%d   sock:%d\n", xSock, sock);

    if( xSock > 0 )
	close( xSock );
    if( sock > 0 )
	close( sock );
}


/* ---------------------------------------------------------
 * CloseClient:
 *
 * --------------------------------------------------------- */

void 
TCPNet::CloseClient( void )
{
    printDebug(4, "CloseClient() xSock:%d\n", xSock);
    if( xSock > 0 )
	close( xSock );
}


/* ---------------------------------------------------------
 *  doSend:
 *  	Send a message on the socket.
 *
 *  	arguments
 *          msg - the message to send
 *          len - the length of the message to be sent
 *
 * --------------------------------------------------------- */

int 
TCPNet::doSend(const char *msg, long len)
{
    long n;
    int err;

    err = pthread_mutex_lock(&smutex);
    if ( err!=0 )
        printDebug( 4, "getRemoteHost--can't lock send mutex\n");

    n = sendto(xSock, msg, len, 0, (struct sockaddr *)&peer, sizeof(peer));
    printDebug( 4, "doSend--sending msg: %s\n",msg);
    err = pthread_mutex_unlock(&smutex);
    if ( err!=0 )
	printDebug( 4, "getRemoteHost--can't unlock send mutex\n");
    if (n < 0)
	perror("TCPNet: sendto()");
    else if (n < len)
       	printf("sendto() didn't send whole message\n");

    return n;
}


/* ---------------------------------------------------------
 *  doRecv:
 *  	Receive a message from the socket.  The receive will block waiting
 *  	for the message the current block_sec and block_usec time.
 *
 *  	arguments
 *          buf - the buffer to copy the  message into
 *          len - the length of the buffer passed in
 *
 * --------------------------------------------------------- */

int TCPNet::doRecv(char *buf, long len)
{
    Socklen_t l;
    int n;
    struct timeval t;
    fd_set fd;

    t.tv_sec = block_secs;
    t.tv_usec = block_usecs;
    FD_ZERO(&fd);
    FD_SET(xSock, &fd);
    n = select(xSock+1, &fd, NULL, NULL, &t);
    if (n > 0){
        l = sizeof(peer);
        n = recvfrom(xSock, buf, len, 0, (struct sockaddr *)&peer, &l);
        printDebug( 4, "doRecv--recvfrom returned %d\n",n);
        if (n < 0)
            printError("doRecv: recvfrom() failed\n");
        else{
             buf[n] = '\0';
             n++;
        }
    }
    if ( block_usecs > 0 || block_secs > 0 )
        printDebug( 4, "doRecv--returning %d\n", n);
    return n;
}


/* ---------------------------------------------------------
 *  doBlockedRecv:
 *  	Receive a message from the socket.  The receive will block waiting
 *  	for the message indefinitely or untill an error occurs on the socket.
 *
 * 	This routine blocks until data arrives on the socket and
 *      extracts the sender's ip addr returned by recvfrom().
 *
 *      buf - the buffer to copy the  message into
 *      len - the length of the buffer passed in
 *      sender - buffer to copy the sender's id into
 *
 * --------------------------------------------------------- */

int TCPNet::doBlockedRecv(char *buf, long len, char *sender)
{
    int n;
    Socklen_t l;
    sender[0] = '\0';  
    l = sizeof(peer);

    n = recvfrom(xSock, buf, len, 0, (struct sockaddr *)&peer, &l);
    if ( errno != EINTR )
        printDebug( 4, "doBlockedRecv--recvfrom returned %d\n",n);
    if (n < 0){  
        if (errno == EINTR){
            printDebug( 4, "recvfrom() returned EINTR\n");
            return -2;
        }
        printDebug( 4, "doBlockedRecv: recvfrom() errno: %d\n", errno);
        printError("TCPNet::doBlockedRecv: recvfrom() failed\n");
    }
    else{
        buf[n] = '\0';
        n++;   
    }
    string* remote_p = getRemoteHost();
    strcpy(sender, remote_p->c_str());
    delete remote_p;
    printDebug( 4, "TCPNet::doBlockedRecv--returning %d\n", n);
    return n;
}
